/*****************************************************************************
 * audio_out.cpp: Symbian Audio Output Module
 *****************************************************************************
 * Copyright © 2010 Pankaj Yadav
 * Author: Pankaj Yadav
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#include <aknnotewrappers.h>
#include <eikenv.h>
#include <eikdef.h>
#include <e32math.h>

extern "C" {

#define bool int
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <vlc_common.h>
#include <vlc_plugin.h>
#include <vlc_aout.h>

}


#include <mda\common\audio.h>
#include <MdaAudioOutputStream.h>
#if !defined(__MDA_COMMON_RESOURCE_H__)
#include <Mda\Common\Resource.h>
#endif

#define FRAME_SIZE 12442


struct aout_sys_t
{
    mtime_t next_date;
    mtime_t buffer_time;
};

typedef struct AudioSpec
{
    int freq;
    TUint16 format;
    TUint8  channels;
    TUint8  silence;
    TUint16 samples;
    TUint16 padding;
    TUint32 size;

    void ( *callback)(void *userdata, short int *stream,unsigned int len);
    void  *userdata;
} AudioSpec;

static int  Open ( vlc_object_t * );
static void Close ( vlc_object_t * );
static void Play ( aout_instance_t * );
static void HardwareCallback ( void *, short int *,unsigned int );
void *createdevice ( void * arg );

vlc_module_begin()
    set_shortname( "SYMBIANAUDIO" )
    set_description( _("Symbian audio output") )
    set_capability( "audio output", 40 )
    set_category( CAT_AUDIO )
    set_subcategory( SUBCAT_AUDIO_AOUT )
    add_shortcut( "sym" )
    set_callbacks( Open, Close )
vlc_module_end()


class MdaSound : public MMdaAudioOutputStreamCallback
{

public:
    int v_freq,v_channels,closed;
    aout_instance_t *v_p_aout;
    CActiveScheduler * scheduler;
    int construct();
    void Open();
    void sclose();

public:
    virtual void MaoscOpenComplete(TInt aError);
    virtual void MaoscBufferCopied(TInt aError, const TDesC8& aBuffer);
    virtual void MaoscPlayComplete(TInt aError);

protected:
    void UpdateBuffer();
    void FillBuffer(aout_instance_t *v_p_aout,TInt16* buffer, TUint len);

protected:
    CMdaAudioOutputStream* iStream;
    TMdaAudioDataSettings iSettings;
    TInt16* iSoundData;
    TPtr8* iSoundBuf;
} SymbianSound;


void MdaSound::sclose()
{
    closed = 1;
}


void MdaSound::FillBuffer(aout_instance_t *v_p_aout,TInt16* buffer, TUint len)
{
    HardwareCallback(v_p_aout,buffer,len);
}


int MdaSound::construct()
{
    iStream = CMdaAudioOutputStream::NewL ( *this );
    if(iStream == NULL)
    {
        return -1;
    }

    iSoundData = new TInt16[FRAME_SIZE];
    if(iSoundData == NULL)
    {
        return -1;
    }

    iSoundBuf = new TPtr8((TUint8*)iSoundData, FRAME_SIZE*2, FRAME_SIZE*2);
    if(iSoundBuf == NULL)
    {
        return -1;
    }

    return 0;
}

void MdaSound::Open()
{
    closed = 0;
    iStream->Open( &iSettings );
}

void MdaSound::MaoscOpenComplete(TInt aError)
{


    if (aError == KErrNone)
    {
        iStream->SetAudioPropertiesL(v_freq, v_channels==1 ?
                                 TMdaAudioDataSettings::EChannelsStereo :TMdaAudioDataSettings::EChannelsMono);
        // note that MaxVolume() is different in the emulator and the real device!
        iStream->SetVolume(iStream->MaxVolume()/2);
        iStream->SetPriority(EPriorityMuchMore, EMdaPriorityPreferenceNone);

        // Fill first buffer and write it to the stream
        UpdateBuffer();
    }
}

void MdaSound::UpdateBuffer()
{
    if(closed == 0)
    {
        iStream->SetVolume((int)(((float)this->v_p_aout->output.i_volume/(float)AOUT_VOLUME_MAX)*((float)iStream->MaxVolume())));
        FillBuffer(v_p_aout,iSoundData, FRAME_SIZE);
        iStream->WriteL(*iSoundBuf);
    }
    else
    {
        iStream->Stop();
    }
}

void MdaSound::MaoscBufferCopied(TInt aError, const TDesC8& )
{
    if (aError == KErrNone)
    {
        UpdateBuffer();
    }
}

void MdaSound::MaoscPlayComplete(TInt aError)
{

    if (aError == KErrUnderflow)
    {
        UpdateBuffer();
        return;
    }

    CActiveScheduler::Stop();

    delete iStream;
    delete iSoundData;
    delete iSoundBuf;
    delete scheduler;
    pthread_exit(NULL);
}

static int Open ( vlc_object_t *p_this )
{

    aout_instance_t *p_aout = (aout_instance_t *)p_this;

    AudioSpec desired, obtained;
    int i_nb_channels;
    pthread_t tid;
    vlc_value_t val, text;

    i_nb_channels = aout_FormatNbChannels( &p_aout->output.output );
    i_nb_channels = 1; /*Temporarily,Stereo has some problem*/

    if(i_nb_channels == 1)
    {
        p_aout->output.output.i_physical_channels = AOUT_CHAN_CENTER;
    }
    else if ( i_nb_channels >= 2 )
    {
        i_nb_channels = 2;
        p_aout->output.output.i_physical_channels = AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT;
    }

    desired.freq       = p_aout->output.output.i_rate;
    desired.format     = 1;
    desired.channels   = i_nb_channels;
    desired.userdata   = p_aout;
    desired.samples    = FRAME_SIZE;
    obtained.channels  = desired.channels;

    p_aout->output.output.i_format = VLC_FOURCC('s','1','6','l');
    p_aout->output.i_nb_samples = FRAME_SIZE;
    p_aout->output.pf_play = Play;
    SymbianSound.v_channels = desired.channels;
    SymbianSound.v_p_aout = p_aout;

    if((desired.freq < 8000) || (desired.freq > 96000))
    {
        return VLC_EGENERIC;
    }
    else if((desired.freq >= 8000) && (desired.freq < 11025))
    {
        SymbianSound.v_freq = 0x10;
    }
    else if((desired.freq >= 11025) && (desired.freq < 12000))
    {
        SymbianSound.v_freq = 0x40;
    }
    else if((desired.freq >= 12000) && (desired.freq < 16000))
    {
        SymbianSound.v_freq = 0x80;
    }
    else if((desired.freq >= 16000) && (desired.freq < 22050))
    {
        SymbianSound.v_freq = 0x100;
    }
    else if((desired.freq >= 22050) && (desired.freq < 24000))
    {
        SymbianSound.v_freq = 0x400;
    }
    else if((desired.freq >= 24000) && (desired.freq < 32000))
    {
        SymbianSound.v_freq = 0x800;
    }
    else if((desired.freq >= 32000) && (desired.freq < 44100))
    {
        SymbianSound.v_freq = 0x1000;
    }
    else if((desired.freq >= 44100) && (desired.freq < 48000))
    {
        SymbianSound.v_freq = 0x4000;
    }
    else if((desired.freq >= 48000) && (desired.freq<64000))
    {
        SymbianSound.v_freq = 0x10000;
    }
    else if((desired.freq >= 64000) && (desired.freq<96000))
    {
        SymbianSound.v_freq = 0x40000;
    }
    else
    {
        SymbianSound.v_freq=0x20000;
    }

    p_aout->output.output.i_rate = SymbianSound.v_freq;

    aout_VolumeSoftInit( p_aout );

    if(pthread_create(&tid,NULL,createdevice,NULL))
    {
        return VLC_EGENERIC;
    }

    if ( obtained.channels != i_nb_channels )
    {
        p_aout->output.output.i_physical_channels = (obtained.channels == 2 ?
                                            AOUT_CHAN_LEFT | AOUT_CHAN_RIGHT :
                                            AOUT_CHAN_CENTER);

        if ( var_Type( p_aout, "audio-device" ) == 0 )
        {
            var_Create( p_aout, "audio-device",
                        VLC_VAR_INTEGER | VLC_VAR_HASCHOICE );
            text.psz_string = _("Audio device");
            var_Change( p_aout, "audio-device", VLC_VAR_SETTEXT, &text, NULL );

            val.i_int = (obtained.channels == 2) ? AOUT_VAR_STEREO :
                        AOUT_VAR_MONO;
            text.psz_string = (obtained.channels == 2) ? (char*)N_("Stereo") :(char*)
                              N_("Mono");
            var_Change( p_aout, "audio-device",
                        VLC_VAR_ADDCHOICE, &val, &text );
            var_AddCallback( p_aout, "audio-device", aout_ChannelsRestart,
                             NULL );
        }
    }
    else if ( var_Type( p_aout, "audio-device" ) == 0 )
    {
        var_Create( p_aout, "audio-device",
                    VLC_VAR_INTEGER | VLC_VAR_HASCHOICE );
        text.psz_string = _("Audio device");
        var_Change( p_aout, "audio-device", VLC_VAR_SETTEXT, &text, NULL );

        val.i_int = AOUT_VAR_STEREO;
        text.psz_string = N_("Stereo");
        var_Change( p_aout, "audio-device", VLC_VAR_ADDCHOICE, &val, &text );
        val.i_int = AOUT_VAR_MONO;
        text.psz_string = N_("Mono");
        var_Change( p_aout, "audio-device", VLC_VAR_ADDCHOICE, &val, &text );
        if ( i_nb_channels == 2 )
        {
            val.i_int = AOUT_VAR_STEREO;
        }
        else
        {
            val.i_int = AOUT_VAR_MONO;
        }
        var_Change( p_aout, "audio-device", VLC_VAR_SETDEFAULT, &val, NULL );
        var_AddCallback( p_aout, "audio-device", aout_ChannelsRestart, NULL );
    }

    val.b_bool = true;
    var_Set( p_aout, "intf-change", val );

    return VLC_SUCCESS;
}

static void Play( aout_instance_t * p_aout )
{
    VLC_UNUSED(p_aout);
}

static void Close ( vlc_object_t *p_this )
{
    SymbianSound.sclose();
}

static void HardwareCallback( void * _p_aout, short int * p_stream,unsigned int i_len )
{
    aout_instance_t * p_aout = (aout_instance_t *)_p_aout;
    aout_buffer_t *   p_buffer;

    vlc_mutex_lock( &p_aout->output_fifo_lock );
    p_buffer = aout_FifoPop( p_aout, &p_aout->output.fifo );
    vlc_mutex_unlock( &p_aout->output_fifo_lock );

    if ( p_buffer != NULL )
    {
        vlc_memcpy( p_stream, p_buffer->p_buffer, 2*i_len );
        aout_BufferFree( p_buffer );
    }
    else
    {
        vlc_memset( p_stream, 0,2*i_len );
    }

}

void *createdevice(void * arg)
{
    if (!CActiveScheduler::Current())
    {
        SymbianSound.scheduler = new (ELeave) CActiveScheduler;
        if(SymbianSound.scheduler)
        {
            CActiveScheduler::Install(SymbianSound.scheduler);
        }
        else
        {
            return NULL;
        }
    }

    if(SymbianSound.construct()!=0)
    {
        return NULL;
    }

    SymbianSound.Open();
    CActiveScheduler::Start();

    return NULL;
}

inline void WriteToBuf(TDes8& aDescBuf, TInt aOffset, TInt aValue)
{
    aDescBuf[aOffset] = (TUint8)(aValue & 0x00ff);
    aDescBuf[aOffset+1] = (TUint8)((aValue & 0xff00)>>8);
}
